from .dialect import Dialect
from .keywords import distinguish_keywords
from .mangler import mangle
from .parser import Statement, parse
from .tokenizer import tokenize


def sqltree(sql: str, dialect: Dialect = Dialect.mysql) -> Statement:
    tokens = tokenize(sql, dialect)
    tokens = distinguish_keywords(tokens, dialect)
    tokens = mangle(tokens)
    return parse(tokens)



from functools import partial
from typing import Sequence

from sqltree.dialect import Dialect
from sqltree.location import Location
from sqltree.tokenizer import Token, TokenType, tokenize


def check(sql: str, tokens: Sequence[Token]) -> None:
    actual = list(tokenize(sql, Dialect.mysql))
    assert actual == tokens


def test_tokenize() -> None:
    sql = "SELECT * FROM table WHERE x = 3 AND y = 'x' AND z = {x} AND alpha = %s -- x"
    L = partial(Location, sql)
    check(
        sql,
        [
            TokenType.identifier.make("SELECT", L(0, 5)),
            TokenType.punctuation.make("*", L(7, 7)),
            TokenType.identifier.make("FROM", L(9, 12)),
            TokenType.identifier.make("table", L(14, 18)),
            TokenType.identifier.make("WHERE", L(20, 24)),
            TokenType.identifier.make("x", L(26, 26)),
            TokenType.punctuation.make("=", L(28, 28)),
            TokenType.number.make("3", L(30, 30)),
            TokenType.identifier.make("AND", L(32, 34)),
            TokenType.identifier.make("y", L(36, 36)),
            TokenType.punctuation.make("=", L(38, 38)),
            TokenType.string.make("'x'", L(40, 42)),
            TokenType.identifier.make("AND", L(44, 46)),
            TokenType.identifier.make("z", L(48, 48)),
            TokenType.punctuation.make("=", L(50, 50)),
            TokenType.placeholder.make("{x}", L(52, 54)),
            TokenType.identifier.make("AND", L(56, 58)),
            TokenType.identifier.make("alpha", L(60, 64)),
            TokenType.punctuation.make("=", L(66, 66)),
            TokenType.placeholder.make("%s", L(68, 69)),
            TokenType.comment.make("-- x", L(71, 74)),
        ],
    )


def test_comment() -> None:
    sql = "SELECT /* c */ * FROM table # x"
    L = partial(Location, sql)
    check(
        sql,
        [
            TokenType.identifier.make("SELECT", L(0, 5)),
            TokenType.comment.make("/* c */", L(7, 13)),
            TokenType.punctuation.make("*", L(15, 15)),
            TokenType.identifier.make("FROM", L(17, 20)),
            TokenType.identifier.make("table", L(22, 26)),
            TokenType.comment.make("# x", L(28, 30)),
        ],
    )



from dataclasses import dataclass, field, replace
from typing import Iterable, Optional, Sequence, Tuple, Union

from sqltree.peeking_iterator import PeekingIterator

from .location import Location
from .tokenizer import Token, TokenType


@dataclass
class EOFError(Exception):
    expected: str

    def __post_init__(self) -> None:
        super().__init__(str(self))

    def __str__(self) -> str:
        return f"Got EOF (expecting {self.expected!r})"


@dataclass
class ParseError(Exception):
    message: str
    location: Location

    def __post_init__(self) -> None:
        super().__init__(self.message)

    @classmethod
    def from_unexpected_token(cls, token: Token, expected: str) -> "ParseError":
        return ParseError(f"Unexpected {token.text!r} (expected {expected})", token.loc)


@dataclass
class Node:
    pass


@dataclass
class Leaf(Node):
    token: Token = field(repr=False, compare=False)


@dataclass
class Keyword(Leaf):
    text: str


@dataclass
class KeywordSequence(Node):
    keywords: Sequence[Keyword]


@dataclass
class Punctuation(Leaf):
    text: str


@dataclass
class Expression(Node):
    pass


@dataclass
class Identifier(Expression, Leaf):
    text: str


@dataclass
class StringLiteral(Expression, Leaf):
    value: str


@dataclass
class IntegerLiteral(Expression, Leaf):
    value: int


@dataclass
class Placeholder(Expression, Leaf):
    text: str


@dataclass
class BinOp(Expression):
    left: Expression
    op: Union[Punctuation, Keyword]
    right: Expression


@dataclass
class Parenthesized(Expression):
    left_punc: Punctuation = field(compare=False, repr=False)
    inner: Expression
    right_punc: Punctuation = field(compare=False, repr=False)


@dataclass
class Star(Expression, Leaf):
    pass


@dataclass
class SelectExpr(Node):
    expr: Expression
    as_kw: Optional[Keyword] = field(compare=False, repr=False)
    alias: Optional[Identifier]
    trailing_comma: Optional[Punctuation] = field(compare=False, repr=False)


@dataclass
class OrderByExpr(Node):
    """Also used for GROUP BY."""

    expr: Expression
    direction_kw: Optional[Keyword] = field(compare=False, repr=False)
    trailing_comma: Optional[Punctuation] = field(compare=False, repr=False)


@dataclass
class Comment(Leaf):
    text: str


@dataclass
class FromClause(Node):
    kw: Keyword = field(compare=False, repr=False)
    table: Expression


@dataclass
class WhereClause(Node):
    kw: Keyword = field(compare=False, repr=False)
    conditions: Expression


@dataclass
class GroupByClause(Node):
    kwseq: KeywordSequence = field(compare=False, repr=False)
    expr: Sequence[OrderByExpr]


@dataclass
class HavingClause(Node):
    kw: Keyword = field(compare=False, repr=False)
    conditions: Expression


@dataclass
class OrderByClause(Node):
    kwseq: KeywordSequence = field(compare=False, repr=False)
    expr: Sequence[OrderByExpr]


@dataclass
class LimitClause(Node):
    # TODO placeholders, offsets
    kw: Keyword = field(compare=False, repr=False)
    row_count: IntegerLiteral


@dataclass
class Statement(Node):
    leading_comments: Sequence[Comment] = field(repr=False)


@dataclass
class Select(Statement):
    select_kw: Keyword = field(compare=False, repr=False)
    select_exprs: Sequence[SelectExpr]
    from_clause: Optional[FromClause] = None
    where: Optional[WhereClause] = None
    group_by: Optional[GroupByClause] = None
    having: Optional[HavingClause] = None
    order_by: Optional[OrderByClause] = None
    # TODO LIMIT


@dataclass
class Delete(Statement):
    delete_kw: Keyword = field(compare=False, repr=False)
    from_clause: FromClause
    where: Optional[WhereClause] = None
    order_by: Optional[OrderByClause] = None
    limit: Optional[LimitClause] = None


@dataclass
class Default(Node):
    kw: Keyword = field(compare=False, repr=False)


@dataclass
class Assignment(Node):
    col_name: Identifier
    eq_punc: Punctuation
    value: Union[Expression, Default]
    trailing_comma: Optional[Punctuation] = None


@dataclass
class SetClause(Node):
    kw: Keyword = field(compare=False, repr=False)
    assignments: Sequence[Assignment]


@dataclass
class Update(Statement):
    update_kw: Keyword = field(compare=False, repr=False)
    table: Expression
    set_clause: SetClause
    where: Optional[WhereClause] = None
    order_by: Optional[OrderByClause] = None
    limit: Optional[LimitClause] = None


def parse(tokens: Iterable[Token]) -> Statement:
    pi = PeekingIterator(list(tokens))
    return _parse_statement(pi)


def _parse_statement(pi: PeekingIterator[Token]) -> Statement:
    first = pi.peek()
    if first is None:
        raise ValueError("SQL is empty")
    elif first.typ is TokenType.keyword:
        if first.text == "SELECT":
            statement = _parse_select(pi)
        elif first.text == "UPDATE":
            statement = _parse_update(pi)
        elif first.text == "DELETE":
            statement = _parse_delete(pi)
        else:
            raise ParseError(f"Unexpected {first.text!r}", first.loc)
        remaining = pi.peek()
        if remaining is not None:
            raise ParseError.from_unexpected_token(remaining, "EOF")
        return statement
    elif first.typ is TokenType.comment:
        pi.next()
        comment = Comment(first, first.text)
        statement = _parse_statement(pi)
        leading_comments = (comment, *statement.leading_comments)
        return replace(statement, leading_comments=leading_comments)
    else:
        raise ParseError(f"Unexpected {first.text!r}", first.loc)


def _parse_from_clause(pi: PeekingIterator[Token]) -> Optional[FromClause]:
    if _next_is_keyword(pi, "FROM"):
        from_kw = _expect_keyword(pi, "FROM")
        table = _parse_expression(pi)
        return FromClause(from_kw, table)
    return None


def _parse_where_clause(pi: PeekingIterator[Token]) -> Optional[WhereClause]:
    if _next_is_keyword(pi, "WHERE"):
        kw = _expect_keyword(pi, "WHERE")
        table = _parse_expression(pi)
        return WhereClause(kw, table)
    return None


def _parse_having_clause(pi: PeekingIterator[Token]) -> Optional[HavingClause]:
    if _next_is_keyword(pi, "HAVING"):
        kw = _expect_keyword(pi, "HAVING")
        table = _parse_expression(pi)
        return HavingClause(kw, table)
    return None


def _parse_group_by_clause(pi: PeekingIterator[Token]) -> Optional[GroupByClause]:
    kwseq = _maybe_consume_keyword_sequence(pi, ["GROUP", "BY"])
    if kwseq is not None:
        exprs = _parse_order_by_list(pi)
        return GroupByClause(kwseq, exprs)
    else:
        return None


def _parse_order_by_clause(pi: PeekingIterator[Token]) -> Optional[OrderByClause]:
    kwseq = _maybe_consume_keyword_sequence(pi, ["ORDER", "BY"])
    if kwseq is not None:
        exprs = _parse_order_by_list(pi)
        return OrderByClause(kwseq, exprs)
    else:
        return None


def _parse_set_clause(pi: PeekingIterator[Token]) -> SetClause:
    kw = _expect_keyword(pi, "SET")
    assignments = []
    while True:
        expr = _parse_assignment(pi)
        assignments.append(expr)
        if expr.trailing_comma is None:
            break
    return SetClause(kw, assignments)


def _parse_limit_clause(pi: PeekingIterator[Token]) -> Optional[LimitClause]:
    if _next_is_keyword(pi, "LIMIT"):
        kw = _expect_keyword(pi, "LIMIT")
        token = _next_or_else(pi, "number")
        if token.typ is not TokenType.number:
            raise ParseError.from_unexpected_token(token, "number")
        return LimitClause(kw, IntegerLiteral(token, int(token.text)))
    return None


def _parse_update(pi: PeekingIterator[Token]) -> Update:
    kw = _expect_keyword(pi, "UPDATE")
    table = _parse_expression(pi)
    set_clause = _parse_set_clause(pi)
    where_clause = _parse_where_clause(pi)
    order_by_clause = _parse_order_by_clause(pi)
    limit_clause = _parse_limit_clause(pi)
    return Update(
        (), kw, table, set_clause, where_clause, order_by_clause, limit_clause
    )


def _parse_delete(pi: PeekingIterator[Token]) -> Delete:
    delete = _expect_keyword(pi, "DELETE")
    from_clause = _parse_from_clause(pi)
    if from_clause is None:
        token = _next_or_else(pi, "FROM")
        raise ParseError.from_unexpected_token(token, "FROM")
    where_clause = _parse_where_clause(pi)
    order_by_clause = _parse_order_by_clause(pi)
    limit_clause = _parse_limit_clause(pi)
    return Delete((), delete, from_clause, where_clause, order_by_clause, limit_clause)


def _parse_select(pi: PeekingIterator[Token]) -> Select:
    select = _expect_keyword(pi, "SELECT")
    select_exprs = []
    while True:
        expr = _parse_select_expr(pi)
        select_exprs.append(expr)
        if expr.trailing_comma is None:
            break

    from_clause = _parse_from_clause(pi)
    where_clause = _parse_where_clause(pi)
    group_by_clause = _parse_group_by_clause(pi)
    having_clause = _parse_having_clause(pi)
    order_by_clause = _parse_order_by_clause(pi)

    return Select(
        (),
        select,
        select_exprs,
        from_clause,
        where_clause,
        group_by_clause,
        having_clause,
        order_by_clause,
    )


def _parse_identifier(pi: PeekingIterator[Token]) -> Identifier:
    token = _next_or_else(pi, "identifier")
    if token.typ is not TokenType.identifier:
        raise ParseError.from_unexpected_token(token, "identifier")
    return Identifier(token, token.text)


def _parse_select_expr(pi: PeekingIterator[Token]) -> SelectExpr:
    expr = _parse_expression(pi)
    as_kw = alias = trailing_comma = None
    if _next_is_keyword(pi, "AS"):
        as_kw = _expect_keyword(pi, "AS")
        alias = _parse_identifier(pi)
    if _next_is_punctuation(pi, ","):
        trailing_comma = _expect_punctuation(pi, ",")
    return SelectExpr(expr, as_kw, alias, trailing_comma)


def _parse_assignment(pi: PeekingIterator[Token]) -> Assignment:
    colname = _parse_identifier(pi)
    punc = _expect_punctuation(pi, "=")
    if _next_is_keyword(pi, "DEFAULT"):
        kw = _expect_keyword(pi, "DEFAULT")
        value = Default(kw)
    else:
        value = _parse_expression(pi)
    if _next_is_punctuation(pi, ","):
        trailing_comma = _expect_punctuation(pi, ",")
    else:
        trailing_comma = None
    return Assignment(colname, punc, value, trailing_comma)


def _parse_order_by_list(pi: PeekingIterator[Token]) -> Sequence[OrderByExpr]:
    exprs = []
    while True:
        expr = _parse_order_by_expr(pi)
        exprs.append(expr)
        if expr.trailing_comma is None:
            return exprs


def _parse_order_by_expr(pi: PeekingIterator[Token]) -> OrderByExpr:
    expr = _parse_expression(pi)
    if _next_is_keyword(pi, "ASC"):
        direction = _expect_keyword(pi, "ASC")
    elif _next_is_keyword(pi, "DESC"):
        direction = _expect_keyword(pi, "DESC")
    else:
        direction = None
    if _next_is_punctuation(pi, ","):
        trailing_comma = _expect_punctuation(pi, ",")
    else:
        trailing_comma = None
    return OrderByExpr(expr, direction, trailing_comma)


def K(text: str) -> Tuple[TokenType, str]:
    return (TokenType.keyword, text)


def P(text: str) -> Tuple[TokenType, str]:
    return (TokenType.punctuation, text)


_BINOP_PRECEDENCE = [
    (P("^"),),
    (P("*"), P("/"), P("DIV"), P("%"), P("MOD")),
    (P("-"), P("+")),
    (P("<<"), P(">>")),
    (P("&"),),
    (P("|"),),
    (
        P("="),
        P("<=>"),
        P(">="),
        P(">"),
        P("<="),
        P("<"),
        P("<>"),
        P("!="),
        K("IS"),
        K("IS NOT"),
        K("NOT LIKE"),
        K("LIKE"),
        K("NOT REGEXP"),
        K("REGEXP"),
        K("NOT IN"),
        K("IN"),
    ),
    # TODO: BETWEEN, CASE, WHEN, THEN, ELSE
    (K("AND"), P("&&")),
    (K("XOR"),),
    (K("OR"), P("||")),
]
_MIN_PRECEDENCE = len(_BINOP_PRECEDENCE) - 1


def _parse_expression(pi: PeekingIterator[Token]) -> Expression:
    return _parse_binop(pi, _MIN_PRECEDENCE)


def _parse_binop(pi: PeekingIterator[Token], precedence: int) -> Expression:
    if precedence == 0:
        left = _parse_simple_expression(pi)
    else:
        left = _parse_binop(pi, precedence - 1)
    while True:
        token = pi.peek()
        if token is None:
            return left
        options = _BINOP_PRECEDENCE[precedence]
        if any(token.typ is typ and token.text == text for typ, text in options):
            pi.next()
            if token.typ is TokenType.punctuation:
                op = Punctuation(token, token.text)
            else:
                assert token.typ is TokenType.keyword
                op = Keyword(token, token.text)
            right = _parse_binop(pi, precedence - 1)
            left = BinOp(left, op, right)
        else:
            return left


def _parse_simple_expression(pi: PeekingIterator[Token]) -> Expression:
    token = _next_or_else(pi, "expression")
    if token.typ is TokenType.punctuation and token.text == "*":
        return Star(token)
    elif token.typ is TokenType.punctuation and token.text == "(":
        inner = _parse_expression(pi)
        right = _expect_punctuation(pi, ")")
        return Parenthesized(Punctuation(token, "("), inner, right)
    elif token.typ is TokenType.identifier:
        return Identifier(token, token.text)
    elif token.typ is TokenType.number:
        return IntegerLiteral(token, int(token.text))
    elif token.typ is TokenType.placeholder:
        return Placeholder(token, token.text)
    elif token.typ is TokenType.string:
        text = token.text[1:-1]
        if token.text[0] == "`":
            return Identifier(token, text)
        else:
            return StringLiteral(token, text)
    else:
        raise ParseError.from_unexpected_token(token, "expression")


def _next_is_punctuation(pi: PeekingIterator[Token], punctuation: str) -> bool:
    token = pi.peek()
    return (
        token is not None
        and token.typ is TokenType.punctuation
        and token.text == punctuation
    )


def _expect_punctuation(pi: PeekingIterator[Token], punctuation: str) -> Punctuation:
    token = _next_or_else(pi, punctuation)
    if token.typ is not TokenType.punctuation or token.text != punctuation:
        raise ParseError.from_unexpected_token(token, repr(punctuation))
    return Punctuation(token, token.text)


def _maybe_consume_keyword_sequence(
    pi: PeekingIterator[Token], keywords: Sequence[str]
) -> Optional[KeywordSequence]:
    keywords_found = []
    for token in pi:
        if (
            token.typ is TokenType.keyword
            and token.text == keywords[len(keywords_found)]
        ):
            keywords_found.append(token)
            if len(keywords_found) == len(keywords):
                return KeywordSequence(
                    [Keyword(token, token.text) for token in keywords_found]
                )
        else:
            break

    for _ in keywords_found:
        pi.wind_back()
    return None


def _next_is_keyword(pi: PeekingIterator[Token], keyword: str) -> bool:
    token = pi.peek()
    return (
        token is not None and token.typ is TokenType.keyword and token.text == keyword
    )


def _expect_keyword(pi: PeekingIterator[Token], keyword: str) -> Keyword:
    token = _next_or_else(pi, keyword)
    if token.typ is not TokenType.keyword or token.text != keyword:
        raise ParseError.from_unexpected_token(token, repr(keyword))
    return Keyword(token, token.text)


def _next_or_else(pi: PeekingIterator[Token], label: str) -> Token:
    try:
        return pi.next()
    except StopIteration:
        raise EOFError(label)



from dataclasses import dataclass
from typing import Generic, Optional, Sequence, TypeVar

T = TypeVar("T")


@dataclass
class PeekingIterator(Generic[T]):
    seq: Sequence[T]
    next_pos: int = 0

    def next(self) -> T:
        if self.has_next():
            self.next_pos += 1
            return self.seq[self.next_pos - 1]
        raise StopIteration

    def __iter__(self) -> "PeekingIterator[T]":
        return self

    def __next__(self) -> T:
        return self.next()

    def wind_back(self) -> None:
        self.next_pos -= 1

    def advance(self) -> None:
        self.next_pos += 1

    def has_next(self) -> bool:
        return self.next_pos < len(self.seq)

    def current(self) -> T:
        return self.seq[self.next_pos - 1]

    def peek(self) -> Optional[T]:
        if self.next_pos < len(self.seq):
            return self.seq[self.next_pos]
        return None



from sqltree.formatter import transform_and_format
from sqltree.parser import BinOp, Punctuation
from sqltree.visitor import Transformer


class FlipConditions(Transformer):
    def visit_BinOp(self, node: BinOp) -> BinOp:
        if node.op.text == "=":
            op = Punctuation(node.op.token, "!=")
        else:
            op = node.op
        return BinOp(node.left, op, node.right)


def test_transform() -> None:
    assert (
        transform_and_format("SELECT * FROM x WHERE a = 3", FlipConditions())
        == "SELECT *\nFROM x\nWHERE a != 3\n"
    )



"""

The mangler mangles the token list to make it easier for the
parser to handle.

Makes the following changes:

- Attaches all comments (except at the beginning of the string) to
  a neighboring token.
- Turns the following pairs of keywords into a single token:
  IS NOT, NOT IN, NOT LIKE, NOT REGEXP

"""
from typing import Iterable

from .location import Location
from .tokenizer import Token, TokenType

KEYWORD_PAIRS = [("IS", "NOT"), ("NOT", "IN"), ("NOT", "LIKE"), ("NOT", "REGEXP")]


def mangle(tokens: Iterable[Token]) -> Iterable[Token]:
    new_tokens = []
    for i, token in enumerate(tokens):
        if i == 0:
            new_tokens.append(token)
            continue
        mangled = False
        for left_kw, right_kw in KEYWORD_PAIRS:
            if _is_keyword(token, right_kw) and _is_keyword(new_tokens[-1], left_kw):
                new_tokens[-1] = _merge_tokens(new_tokens[-1], token)
                mangled = True
                break
        if mangled:
            continue
        if token.typ is TokenType.comment:
            old_token = new_tokens[-1]
            new_tokens[-1] = Token(
                old_token.typ,
                old_token.text,
                old_token.loc,
                [*old_token.comments, token],
            )
            continue
        new_tokens.append(token)
    return new_tokens


def _is_keyword(token: Token, keyword: str) -> bool:
    return token.typ is TokenType.keyword and token.text == keyword


def _merge_tokens(left: Token, right: Token) -> Token:
    loc = Location(left.loc.sql, left.loc.start_index, right.loc.end_index)
    return Token(
        TokenType.keyword,
        f"{left.text} {right.text}",
        loc,
        [*left.comments, *right.comments],
    )



import collections.abc
from dataclasses import fields
from typing import Generic, Optional, TypeVar

from .parser import Node

T = TypeVar("T")


class Visitor(Generic[T]):
    def visit(self, node: Node) -> T:
        method_name = f"visit_{type(node).__name__}"
        method = getattr(self, method_name, self.generic_visit)
        return method(node)  # type: ignore

    def maybe_visit(self, node: Optional[Node]) -> Optional[T]:
        if node is None:
            return None
        return self.visit(node)

    def generic_visit(self, node: Node) -> T:
        raise NotImplementedError(node)


class Transformer(Visitor[Node]):
    def generic_visit(self, node: Node) -> Node:
        cls = type(node)
        kwargs = {}
        for field in fields(node):
            key = field.name
            value = getattr(node, key)
            if isinstance(value, Node):
                kwargs[key] = self.visit(value)
            elif isinstance(value, collections.abc.Sequence) and not isinstance(
                value, str
            ):
                kwargs[key] = [self.visit(member) for member in value]
            else:
                kwargs[key] = value
        return cls(**kwargs)



from sqltree.formatter import format


def test_select() -> None:
    assert format("SELECT * from x") == "SELECT *\nFROM x\n"
    assert (
        format("# comment\nSELECT * from x where -- comment\n x=3")
        == "# comment\nSELECT *\nFROM x\nWHERE -- comment\nx = 3\n"
    )
    assert format("SELECT x not -- hello\nLIKE y") == "SELECT x NOT LIKE -- hello\ny\n"
    assert (
        format(
            "select x as   z from y group  -- hello\n  by x asc,y desc having x={x}"
            " order  by x desc"
        )
        == "SELECT x AS z\nFROM y\nGROUP BY -- hello\nx ASC, y DESC\nHAVING x ="
        " {x}\nORDER"
        " BY x DESC\n"
    )


def test_update() -> None:
    assert (
        format("update x set y = default, z =3 where x=4 order   by z limit 1")
        == "UPDATE x\nSET y = DEFAULT, z = 3\nWHERE x = 4\nORDER BY z\nLIMIT 1\n"
    )


def test_delete() -> None:
    assert (
        format("delete from x where y = 3 order by z desc limit 1")
        == "DELETE FROM x\nWHERE y = 3\nORDER BY z DESC\nLIMIT 1\n"
    )



from typing import Dict, Iterable, Set

from .dialect import Dialect
from .tokenizer import Token, TokenType

# from https://dev.mysql.com/doc/refman/5.7/en/keywords.html#keywords-in-current-series
MYSQL_KEYWORDS = {
    "ACCESSIBLE",
    "ADD",
    "ALL",
    "ALTER",
    "ANALYZE",
    "AND",
    "AS",
    "ASC",
    "ASENSITIVE",
    "BEFORE",
    "BETWEEN",
    "BIGINT",
    "BINARY",
    "BLOB",
    "BOTH",
    "BY",
    "CALL",
    "CASCADE",
    "CASE",
    "CHANGE",
    "CHAR",
    "CHARACTER",
    "CHECK",
    "COLLATE",
    "COLUMN",
    "CONDITION",
    "CONSTRAINT",
    "CONTINUE",
    "CONVERT",
    "CREATE",
    "CROSS",
    "CURRENT_DATE",
    "CURRENT_TIME",
    "CURRENT_TIMESTAMP",
    "CURRENT_USER",
    "CURSOR",
    "DATABASE",
    "DATABASES",
    "DAY_HOUR",
    "DAY_MICROSECOND",
    "DAY_MINUTE",
    "DAY_SECOND",
    "DEC",
    "DECIMAL",
    "DECLARE",
    "DEFAULT",
    "DELAYED",
    "DELETE",
    "DESC",
    "DESCRIBE",
    "DETERMINISTIC",
    "DISTINCT",
    "DISTINCTROW",
    "DIV",
    "DOUBLE",
    "DROP",
    "DUAL",
    "EACH",
    "ELSE",
    "ELSEIF",
    "ENCLOSED",
    "ESCAPED",
    "EXISTS",
    "EXIT",
    "EXPLAIN",
    "FALSE",
    "FETCH",
    "FLOAT",
    "FLOAT4",
    "FLOAT8",
    "FOR",
    "FORCE",
    "FOREIGN",
    "FROM",
    "FULLTEXT",
    "GENERATED",
    "GET",
    "GRANT",
    "GROUP",
    "HAVING",
    "HIGH_PRIORITY",
    "HOUR_MICROSECOND",
    "HOUR_MINUTE",
    "HOUR_SECOND",
    "IF",
    "IGNORE",
    "IN",
    "INDEX",
    "INFILE",
    "INNER",
    "INOUT",
    "INSENSITIVE",
    "INSERT",
    "INT",
    "INT1",
    "INT2",
    "INT3",
    "INT4",
    "INT8",
    "INTEGER",
    "INTERVAL",
    "INTO",
    "IO_AFTER_GTIDS",
    "IO_BEFORE_GTIDS",
    "IS",
    "ITERATE",
    "JOIN",
    "KEY",
    "KEYS",
    "KILL",
    "LEADING",
    "LEAVE",
    "LEFT",
    "LIKE",
    "LIMIT",
    "LINEAR",
    "LINES",
    "LOAD",
    "LOCALTIME",
    "LOCALTIMESTAMP",
    "LOCK",
    "LONG",
    "LONGBLOB",
    "LONGTEXT",
    "LOOP",
    "LOW_PRIORITY",
    "MASTER_BIND",
    "MASTER_SSL_VERIFY_SERVER_CERT",
    "MATCH",
    "MAXVALUE",
    "MEDIUMBLOB",
    "MEDIUMINT",
    "MEDIUMTEXT",
    "MIDDLEINT",
    "MINUTE_MICROSECOND",
    "MINUTE_SECOND",
    "MOD",
    "MODIFIES",
    "NATURAL",
    "NOT",
    "NO_WRITE_TO_BINLOG",
    "NULL",
    "NUMERIC",
    "ON",
    "OPTIMIZE",
    "OPTIMIZER_COSTS",
    "OPTION",
    "OPTIONALLY",
    "OR",
    "ORDER",
    "OUT",
    "OUTER",
    "OUTFILE",
    "PARTITION",
    "PRECISION",
    "PRIMARY",
    "PROCEDURE",
    "PURGE",
    "RANGE",
    "READ",
    "READS",
    "READ_WRITE",
    "REAL",
    "REFERENCES",
    "REGEXP",
    "RELEASE",
    "RENAME",
    "REPEAT",
    "REPLACE",
    "REQUIRE",
    "RESIGNAL",
    "RESTRICT",
    "RETURN",
    "REVOKE",
    "RIGHT",
    "RLIKE",
    "SCHEMA",
    "SCHEMAS",
    "SECOND_MICROSECOND",
    "SELECT",
    "SENSITIVE",
    "SEPARATOR",
    "SET",
    "SHOW",
    "SIGNAL",
    "SMALLINT",
    "SPATIAL",
    "SPECIFIC",
    "SQL",
    "SQLEXCEPTION",
    "SQLSTATE",
    "SQLWARNING",
    "SQL_BIG_RESULT",
    "SQL_CALC_FOUND_ROWS",
    "SQL_SMALL_RESULT",
    "SSL",
    "STARTING",
    "STORED",
    "STRAIGHT_JOIN",
    "TABLE",
    "TERMINATED",
    "THEN",
    "TINYBLOB",
    "TINYINT",
    "TINYTEXT",
    "TO",
    "TRAILING",
    "TRIGGER",
    "TRUE",
    "UNDO",
    "UNION",
    "UNIQUE",
    "UNLOCK",
    "UNSIGNED",
    "UPDATE",
    "USAGE",
    "USE",
    "USING",
    "UTC_DATE",
    "UTC_TIME",
    "UTC_TIMESTAMP",
    "VALUES",
    "VARBINARY",
    "VARCHAR",
    "VARCHARACTER",
    "VARYING",
    "VIRTUAL",
    "WHEN",
    "WHERE",
    "WHILE",
    "WITH",
    "WRITE",
    "XOR",
    "YEAR_MONTH",
    "ZEROFILL",
}

KEYWORDS: Dict[Dialect, Set[str]] = {Dialect.mysql: MYSQL_KEYWORDS}


def distinguish_keywords(tokens: Iterable[Token], dialect: Dialect) -> Iterable[Token]:
    for token in tokens:
        if token.typ is TokenType.identifier:
            text = token.text.upper()
            if text in KEYWORDS[dialect]:
                yield TokenType.keyword.make(text, token.loc)
            else:
                yield token
        else:
            yield token



from sqltree import parser as p
from sqltree.location import Location
from sqltree.sqltree import sqltree
from sqltree.tokenizer import Token, TokenType

T = Token(TokenType.keyword, "x", Location("x", 1, 1))


def test() -> None:
    tree = sqltree("SELECT * FROM a")
    assert tree == p.Select(
        (),
        p.Keyword(T, "SELECT"),
        [p.SelectExpr(p.Star(T), None, None, None)],
        p.FromClause(p.Keyword(T, "FROM"), p.Identifier(T, "a")),
    )

    tree = sqltree(
        "SELECT a, b AS c FROM a WHERE x = 3 * 3 + 2 * 4 AND y = 'x' AND z NOT IN {x}"
    )
    assert tree == p.Select(
        (),
        p.Keyword(T, "SELECT"),
        [
            p.SelectExpr(p.Identifier(T, "a"), None, None, p.Punctuation(T, ",")),
            p.SelectExpr(
                p.Identifier(T, "b"), p.Keyword(T, "AS"), p.Identifier(T, "c"), None
            ),
        ],
        p.FromClause(p.Keyword(T, "FROM"), p.Identifier(T, "a")),
        p.WhereClause(
            p.Keyword(T, "WHERE"),
            p.BinOp(
                p.BinOp(
                    p.BinOp(
                        p.Identifier(T, "x"),
                        p.Punctuation(T, "="),
                        p.BinOp(
                            p.BinOp(
                                p.IntegerLiteral(T, 3),
                                p.Punctuation(T, "*"),
                                p.IntegerLiteral(T, 3),
                            ),
                            p.Punctuation(T, "+"),
                            p.BinOp(
                                p.IntegerLiteral(T, 2),
                                p.Punctuation(T, "*"),
                                p.IntegerLiteral(T, 4),
                            ),
                        ),
                    ),
                    p.Keyword(T, "AND"),
                    p.BinOp(
                        p.Identifier(T, "y"),
                        p.Punctuation(T, "="),
                        p.StringLiteral(T, "x"),
                    ),
                ),
                p.Keyword(T, "AND"),
                p.BinOp(
                    p.Identifier(T, "z"),
                    p.Keyword(T, "NOT IN"),
                    p.Placeholder(T, "{x}"),
                ),
            ),
        ),
    )

    tree = sqltree("-- comment\nSELECT * FROM a")
    assert tree == p.Select(
        (p.Comment(T, "-- comment\n"),),
        p.Keyword(T, "SELECT"),
        [p.SelectExpr(p.Star(T), None, None, None)],
        p.FromClause(p.Keyword(T, "FROM"), p.Identifier(T, "a")),
    )



import enum
from collections import defaultdict
from dataclasses import dataclass
from typing import Dict, Iterable, Sequence, Set

from .dialect import Dialect
from .location import Location
from .peeking_iterator import PeekingIterator


class TokenizeError(Exception):
    pass


class TokenType(enum.Enum):
    keyword = 1
    punctuation = 2
    string = 3
    number = 4
    identifier = 5
    placeholder = 6
    comment = 7

    def make(self, text: str, loc: Location) -> "Token":
        return Token(self, text, loc)


@dataclass
class Token:
    typ: TokenType
    text: str
    loc: Location
    comments: Sequence["Token"] = ()


PUNCTUATION = {
    ".",
    "(",
    ")",
    ",",
    "+",
    "*",
    ">=",
    "<=",
    "=",
    "<>",
    "!=",
    "/",
    "%",
    "-",
    "~",
    "&",
    "^",
    "|",
    ">>",
    "<<",
    "&&",
    "||",
    "--",  # Not a punctuation but a comment
    "/*",  # Also a comment
}
QUOTATIONS = {"`", "'", '"'}


def tokenize(sql: str, dialect: Dialect) -> Iterable[Token]:
    # Prepare punctuation
    starting_char_to_continuations: Dict[str, Set[str]] = defaultdict(set)
    for punc in PUNCTUATION:
        if len(punc) == 1:
            starting_char_to_continuations[punc].add("")
        elif len(punc) == 2:
            starting_char_to_continuations[punc[0]].add(punc[1])
        else:
            raise TokenizeError(f"don't know how to handle {punc}")

    pi = PeekingIterator(sql)
    while pi.has_next():
        start_index = pi.next_pos
        char = pi.next()
        if char.isalpha():
            pi.wind_back()
            text = _consume_identifier(pi)
            token_type = TokenType.identifier
        elif char.isspace():
            continue  # Skip over whitespace
        elif char == "%":
            next_char = pi.peek()
            if next_char is not None and next_char.isalpha():
                token_type = TokenType.placeholder
                text = "%" + _consume_identifier(pi)
            else:
                token_type = TokenType.punctuation
                text = "%"
        elif char in starting_char_to_continuations:
            token_type = TokenType.punctuation
            continuations = starting_char_to_continuations[char]
            if not pi.has_next():
                if "" in continuations:
                    text = char
                else:
                    raise TokenizeError(
                        f"unexpected EOF following {char} (expected one of"
                        f" {continuations})"
                    )
            else:
                c = pi.next()
                if c in continuations:
                    text = char + c
                    if text == "--":
                        token_type = TokenType.comment
                        text += _consume_until(pi, "\n", eof_okay=True)
                    elif text == "/*":
                        token_type = TokenType.comment
                        chars = []
                        seen_star = False
                        for c in pi:
                            chars.append(c)
                            if seen_star and c == "/":
                                text += "".join(chars)
                                break
                            if c == "*":
                                seen_star = True
                            else:
                                seen_star = False
                        else:
                            raise TokenizeError("unexpected EOF (expected '*/')")
                elif "" in c:
                    pi.wind_back()
                    text = char
                else:
                    raise TokenizeError(
                        f"unexpected {c} following {char} (expected one of"
                        f" {continuations})"
                    )
        elif char.isnumeric():
            # TODO floats, hex, other kinds of numbers?
            pi.wind_back()
            token_type = TokenType.number
            text = _consume_integer(pi)
        elif char in QUOTATIONS:
            token_type = TokenType.string
            text = char + _consume_until(pi, char)
        elif char == "{":
            token_type = TokenType.placeholder
            text = "{" + _consume_until(pi, "}")
        elif char == "#":
            token_type = TokenType.comment
            text = "#" + _consume_until(pi, "\n", eof_okay=True)
        else:
            # TODO comments. Maybe we leave those out of the AST in the parser, then reattach
            # them later at the nearest node we find. Or we just attach them to the token and make
            # sure each token object stays in the AST.
            raise TokenizeError(f"unexpected character {char}")
        yield Token(token_type, text, Location(sql, start_index, pi.next_pos - 1))


def _consume_until(pi: PeekingIterator[str], end: str, eof_okay: bool = False) -> str:
    chars = []
    for c in pi:
        chars.append(c)
        # TODO backslash escapes?
        if c == end:
            return "".join(chars)
    if eof_okay:
        return "".join(chars)
    raise TokenizeError(f"unexpected EOF (expected {end!r})")


def _consume_identifier(pi: PeekingIterator[str]) -> str:
    chars = []
    for c in pi:
        if c.isalnum() or c == "_":
            chars.append(c)
        else:
            pi.wind_back()
            return "".join(chars)
    return "".join(chars)


def _consume_integer(pi: PeekingIterator[str]) -> str:
    chars = []
    for c in pi:
        if c.isnumeric():
            chars.append(c)
        else:
            pi.wind_back()
            return "".join(chars)
    return "".join(chars)



import argparse
from contextlib import contextmanager
from dataclasses import dataclass, field
from typing import Generator, List, Sequence

from . import parser as p
from .sqltree import sqltree
from .tokenizer import Token
from .visitor import Transformer, Visitor


@dataclass
class Formatter(Visitor[None]):
    pieces: List[str] = field(default_factory=list)
    should_skip_comments: bool = False

    def add_space(self) -> None:
        if self.pieces and not self.pieces[-1].endswith("\n"):
            self.pieces.append(" ")

    def add_comments(self, comments: Sequence[Token]) -> None:
        if comments:
            self.add_space()
        for comment in comments:
            self.pieces.append(comment.text)

    @contextmanager
    def skip_comments(self) -> Generator[None, None, None]:
        old_value = self.should_skip_comments
        try:
            self.should_skip_comments = True
            yield
        finally:
            self.should_skip_comments = old_value

    def visit(self, node: p.Node) -> None:
        if isinstance(node, p.Statement):
            for comment in node.leading_comments:
                self.pieces.append(comment.text)
        super().visit(node)
        if not self.should_skip_comments and isinstance(node, p.Leaf):
            self.add_comments(node.token.comments)

    def visit_KeywordSequence(self, node: p.KeywordSequence) -> None:
        # Move all the comments to the end
        with self.skip_comments():
            for i, kw in enumerate(node.keywords):
                if i > 0:
                    self.add_space()
                self.visit(kw)
        for kw in node.keywords:
            self.add_comments(kw.token.comments)

    def visit_FromClause(self, node: p.FromClause) -> None:
        self.visit(node.kw)
        self.add_space()
        self.visit(node.table)
        self.pieces.append("\n")

    def visit_WhereClause(self, node: p.WhereClause) -> None:
        self.visit(node.kw)
        self.add_space()
        self.visit(node.conditions)
        self.pieces.append("\n")

    def visit_HavingClause(self, node: p.HavingClause) -> None:
        self.visit(node.kw)
        self.add_space()
        self.visit(node.conditions)
        self.pieces.append("\n")

    def visit_GroupByClause(self, node: p.GroupByClause) -> None:
        self.visit(node.kwseq)
        self.add_space()
        for i, expr in enumerate(node.expr):
            if i > 0:
                self.add_space()
            self.visit(expr)
        self.pieces.append("\n")

    def visit_OrderByClause(self, node: p.OrderByClause) -> None:
        self.visit(node.kwseq)
        self.add_space()
        for i, expr in enumerate(node.expr):
            if i > 0:
                self.add_space()
            self.visit(expr)
        self.pieces.append("\n")

    def visit_SetClause(self, node: p.SetClause) -> None:
        self.visit(node.kw)
        self.add_space()
        for i, assignment in enumerate(node.assignments):
            if i > 0:
                self.add_space()
            self.visit(assignment)
        self.pieces.append("\n")

    def visit_Assignment(self, node: p.Assignment) -> None:
        self.visit(node.col_name)
        self.add_space()
        self.visit(node.eq_punc)
        self.add_space()
        self.visit(node.value)
        if node.trailing_comma:
            self.visit(node.trailing_comma)

    def visit_Default(self, node: p.Default) -> None:
        self.visit(node.kw)

    def visit_LimitClause(self, node: p.LimitClause) -> None:
        self.visit(node.kw)
        self.add_space()
        self.visit(node.row_count)
        self.pieces.append("\n")

    def visit_Select(self, node: p.Select) -> None:
        self.visit(node.select_kw)
        self.add_space()
        for i, expr in enumerate(node.select_exprs):
            if i > 0:
                self.add_space()
            self.visit(expr)
        self.pieces.append("\n")

        self.maybe_visit(node.from_clause)
        self.maybe_visit(node.where)
        self.maybe_visit(node.group_by)
        self.maybe_visit(node.having)
        self.maybe_visit(node.order_by)

    def visit_Delete(self, node: p.Delete) -> None:
        self.visit(node.delete_kw)
        self.add_space()
        self.visit(node.from_clause)
        self.maybe_visit(node.where)
        self.maybe_visit(node.order_by)
        self.maybe_visit(node.limit)

    def visit_Update(self, node: p.Update) -> None:
        self.visit(node.update_kw)
        self.add_space()
        self.visit(node.table)
        self.pieces.append("\n")
        self.visit(node.set_clause)
        self.maybe_visit(node.where)
        self.maybe_visit(node.order_by)
        self.maybe_visit(node.limit)

    def visit_Keyword(self, node: p.Keyword) -> None:
        self.pieces.append(node.text)

    def visit_Punctuation(self, node: p.Punctuation) -> None:
        self.pieces.append(node.text)

    def visit_Identifier(self, node: p.Identifier) -> None:
        self.pieces.append(node.text)

    def visit_Placeholder(self, node: p.Placeholder) -> None:
        self.pieces.append(node.text)

    def visit_IntegerLiteral(self, node: p.IntegerLiteral) -> None:
        self.pieces.append(str(node.value))

    def visit_StringLiteral(self, node: p.StringLiteral) -> None:
        self.pieces.append(f'"{node.value}"')

    def visit_Star(self, node: p.Star) -> None:
        self.pieces.append("*")

    def visit_Parenthesized(self, node: p.Parenthesized) -> None:
        self.visit(node.left_punc)
        self.visit(node.inner)
        self.visit(node.right_punc)

    def visit_BinOp(self, node: p.BinOp) -> None:
        self.visit(node.left)
        self.add_space()
        self.visit(node.op)
        self.add_space()
        self.visit(node.right)

    def visit_SelectExpr(self, node: p.SelectExpr) -> None:
        self.visit(node.expr)
        if node.as_kw is not None and node.alias is not None:
            self.add_space()
            self.visit(node.as_kw)
            self.add_space()
            self.visit(node.alias)
        if node.trailing_comma is not None:
            self.visit(node.trailing_comma)

    def visit_OrderByExpr(self, node: p.OrderByExpr) -> None:
        self.visit(node.expr)
        if node.direction_kw is not None:
            self.add_space()
            self.visit(node.direction_kw)
        if node.trailing_comma is not None:
            self.visit(node.trailing_comma)


def format_tree(tree: p.Node) -> str:
    fmt = Formatter()
    fmt.visit(tree)
    return "".join(fmt.pieces)


def format(sql: str) -> str:
    return format_tree(sqltree(sql))


def transform_and_format(sql: str, transformer: Transformer) -> str:
    tree = sqltree(sql)
    new_tree = transformer.visit(tree)
    return format_tree(new_tree)


if __name__ == "__main__":
    parser = argparse.ArgumentParser("sqltree")
    parser.add_argument("sql", help="SQL string to format")
    args = parser.parse_args()
    print(format(args.sql), end="")



import argparse

from .sqltree import sqltree

if __name__ == "__main__":
    parser = argparse.ArgumentParser("sqltree")
    parser.add_argument("sql", help="SQL string to parse")
    args = parser.parse_args()
    print(sqltree(args.sql))
